/*
 * jedec2.c
 */

#include <stdlib.h>

#include "jedec2.h"
#include "config2.h"

//#include "galasm.h"


/* input:   aBuffer     Pointer to zero-terminated text-buffer to checksum */
/* output:  16 bit unsigned checksum of the JEDEC structure within <STX> and <ETX> */
uint16_t fileChecksum(uint8_t *aBuffer)
{
	uint16_t	checksum;
	uint8_t		c;

	checksum = 0;

	s = aBuffer;
	while(1)
	{
		c = *s++;
		if(0x00 == c)
		{
			break;
		}
		if(STX == c)			/* <STX> is included in checksum */
		{
			while(c && ETX != c)
			{
				checksum += c;
				c = *s++;
			}
			checksum += c;		/* <ETX> is included in checksum */
			break;			/* done, exit loop */
		}
	}
	return(checksum);
}

uint16_t numberOfFuses(uint16_t aGalType)
{
	switch(aGalType)
	{
	  case GAL16V8:
		return(NUMOFFUSES16);
	  case GAL20V8:
		return(NUMOFFUSES20);
	  case GAL22V10:
		return(NUMOFFUSES22V10);
	  case GAL20RA10:
		return(NUMOFFUSES20RA10);
	}
	return(0);
}

uint16_t fuseChecksum(struct JedecStruct *aJedec)
{
	uint16_t	checksum;
	uint16_t	byte;
	uint16_t	n;
	uint16_t	xorPos;
	uint16_t	sigPos;
	uint16_t	endPos;
	uint8_t		*p;
	uint8_t		*q;

	p = aJedec->GALLogic;
	q = NULL;

	n = 0;
	checksum = 0;
	byte = 0;

	endPos = numberOfFuses(aGalType);
	switch(aGalType)
	{
	  case GAL16V8:
		xorPos = XOR16;
		sigPos = SIG16;
		break;
	  case GAL20V8:
		xorPos = XOR20;
		sigPos = SIG20;
		break;
	  case GAL22V10:
		xorPos = XOR22V10;
		sigPos = SIG22V10;
		break;
	  case GAL20RA10:
		xorPos = XOR20RA10;
		sigPos = SIG20RA10;
		break;
	}

	while(n < endPos)
	{
		if(xorPos == n)
		{
			p = aJedec->GALXOR;
			if(GAL22V10 == aGalType)
			{
				q = aJedec->GALS1;
			}
		}
		else if(sigPos == n)
		{
			p = aJedec->GALSig;
			q = NULL;
		}

		byte |= *p++ << (n & 7);	/* insert bit */
		n++;
		if(0 == (n & 7))
		{
			checksum += byte;
			byte = 0;
		}
		if(q)
		{
			byte |= *p++ << (n & 7);
			n++;
			if(0 == (n & 7))
			{
				checksum += byte;
				byte = 0;
			}
		}
	}
}

uint8_t *putString(uint8_t *aBuffer, const char *aString)
{
	const uint8_t	*s;
	uint8_t		*d;
	uint8_t		c;

	d = aBuffer;
	if(d)
	{
		s = (uint8_t *)aString;
		if(s)
		{
			c = *s++;
			while(c)
			{
				*d++ = c;
				c = *s++;
			}
		}
		*d = '\0';
	}
	return(d);
}

uint8_t *putLine(uint8_t *aBuffer, const char *aString)
{
	uint8_t	*d;

	d = putString(aBuffer, aString);
	if(d)
	{
		*d++ = LF;
		*d = '\0';
	}
	return(d);
}

uint8_t *putDecWord(uint8_t *aBuffer, uint16_t aValue)
{
	uint8_t		buf[8];
	uint8_t		*d;

	d = &buf[7];
	*d = '\0';

	while(1)
	{
		*--d = '0' + (aValue % 10);
		aValue = aValue / 10;
		if(0 == aValue)
		{
			break;
		}
	}
	return(putString(aBuffer, d));
}

uint8_t *putBits(uint8_t *aBuffer, uint8_t *aBits, uint16_t aAddress, uint16_t aLength, uint8_t aDefaultBit)
{
	uint8_t	*d;
	uint16_t	i;

	d = aBuffer;
	if(d)
	{
		*d++ = 'L';
		d = putDecWord(d, aAddress);
		*d++ = ' ';
		for(i = 0; i <= aLength; i++)
		{
			c = *aBits++;
			if(aDefaultBit != c)
			{
				aBuffer = NULL;
			}
			*d++ = '0' + c;
		}
		*d++ = '*';
		*d++ = LF;
		if(aBuffer)
		{
			d = aBuffer;
		}
		*d = '\0';
	}
	return(d);
}

uint16_t pure(uint8_t *aBuffer, uint16_t aLength, uint8_t aValue)
{
	uint8_t	*s;
	uint16_t	count;

	s = aBuffer;
	while(aLength--)
	{
		if(aValue != *s++)
		{
			return(0);
		}
	}
	return(s - aBuffer);
}

void writeJedecFile(const char *aPathname, int aGalType, struct JedecStruct *aJedec, struct Config *aConfiguration)
{
	uint8_t		*buffer;
	uint8_t		*p;
	uint8_t		*q;
	const uint8_t	*s;
	uint32_t	length;
	uint8_t		defaultBit;
	uint8_t		c;
	uint8_t		xorBuf[32];
	uint16_t	ones;
	uint16_t	zeros;
	uint16_t	xorLength;

	buffer = (uint8_t *) malloc(65536);
	if(buffer)
	{
		ones = 0;
		zeros = 0;

		p = xorBuf;
		for(i = 0; i < XORSize; i++)
		{
			*p++ = aJedec->GALXOR[i];
			if(GAL22V10 == aGalType)
			{
				*p++ = aJedec->GALS1[i];
			}
		}
		xorLength = p - xorBuf;
		zeros += pure(xorBuf, XORSize, 0);
		ones += pure(xorBuf, XORSize, 1);

		for(r = 0; r < rowSize; r++)
		{
			zeros += pure(&aJedec->GALLogic[address], maxFuseAddress, 0);
			ones += pure(&aJedec->GALLogic[address], maxFuseAddress, 0);
			address += maxFuseAddress;
		}

		defaultBit = ones > zeros ? 1 : 0;

		p = buffer;

		*p++ = STX;
		*p++ = LF;

		p = putLine(p, "Used Program:   GALasm " GALASM_VERSION);
		p = putLine(p, "GAL-Assembler:  GALasm " GALASM_VERSION);
		p = putString(p, "Device:         GAL");
		p = putLine(p, getGALName(aGalType));

		*p++ = 'F';
		*p++ = '0' + defaultBit;				/* default fuse value */
		*p++ = '*';
		*p++ = LF;

		*p++ = 'G';
		*p++ = aConfiguration->JedecSecBit ? '1' : '0';		/* security bit */
		*p++ = '*';
		*p++ = LF;

		*p++ = 'Q';
		*p++ = 'F';
		p = putDecWord(p, numberOfFuses(aGalType));
		*p++ = '*';
		*p++ = LF;

		address = 0;
		for(r = 0; r < rowSize; r++)
		{
			p = putBits(p, &aJedec->GALLogic[address], address, maxFuseAddress, defaultBit);
			address += maxFuseAddress;
		}

		p = putBits(p, xorBuf, address, xorLength, defaultBit);
		address += xorLength;

		p = putBits(p, aJedec->GALSig, address, SIG_SIZE, defaultBit);
		address += SIG_SIZE;

		if(GAL16V8 == aGalType || GAL20V8 == aGalType)
		{
			p = putBits(p, aJedec->GALAC1, address, AC1_SIZE, defaultBit);
			address += AC1_SIZE;

			p = putBits(p, aJedec->GALPT, address, PT_SIZE, defaultBit);
			address += PT_SIZE;

			p = putBits(p, &aJedec->GALSYN, address, 1, defaultBit);
			address += 1;

			p = putBits(p, &aJedec->GALAC0, address, 1, defaultBit);
			address += 1;
		}

		*p++ = 'C';
		p = putDecWord(p, fuseChecksum(aGalType));
		*p++ = '*';
		*p++ = LF;

		*p++ = '*';
		*p++ = LF;

		*p++ = ETX;
		*p++ = LF;

		*p = '\0';

		p = putDecWord(p, fileChecksum(p));

		length = p - buffer;

		f_save(aPathname, buffer, length);
		free(buffer);
	}
	else
	{
		ErrorReq(2);		/* out of memory */
	}
}

